#include "include.h"

using namespace std;

// globals
rgb_t vram[MNIST_IMG_WIDTH * MNIST_IMG_HEIGHT]; // hold image buffer (pixels RGBA)
unsigned int cxClient;
unsigned int cyClient;
#define WM_TICK 1

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

void set_pixel(int x, int y, rgb_t value)
{
	if (x >= MNIST_IMG_WIDTH || x < 0)
		return;
	if (y >= MNIST_IMG_HEIGHT || y < 0)
		return;

	vram[x + MNIST_IMG_WIDTH * y].r = value.r;
	vram[x + MNIST_IMG_WIDTH * y].g = value.g;
	vram[x + MNIST_IMG_WIDTH * y].b = value.b;
}

void add_pixel(int x, int y, rgb_t value)
{
	if (x >= MNIST_IMG_WIDTH || x < 0)
		return;
	if (y >= MNIST_IMG_WIDTH || y < 0)
		return;

	if (vram[x + MNIST_IMG_WIDTH * y].r == 0 &&
		vram[x + MNIST_IMG_WIDTH * y].g == 0 &&
		vram[x + MNIST_IMG_WIDTH * y].b == 0)
	{
		return;
	}

	int r = vram[x + MNIST_IMG_WIDTH * y].r;
	int g = vram[x + MNIST_IMG_WIDTH * y].g;
	int b = vram[x + MNIST_IMG_WIDTH * y].b;

	r -= value.r;
	g -= value.g;
	b -= value.b;

	if (r < 0)
		r = 0;
	if (g < 0)
		g = 0;
	if (b < 0)
		b = 0;

	vram[x + MNIST_IMG_WIDTH * y].r = r;
	vram[x + MNIST_IMG_WIDTH * y].g = g;
	vram[x + MNIST_IMG_WIDTH * y].b = b;

}

void clear_pixels()
{
	memset(vram, ~0, MNIST_IMG_WIDTH * MNIST_IMG_HEIGHT * sizeof(int));
}

uint32_t flipBytes(uint32_t n)
{
	uint32_t b0, b1, b2, b3;

	b0 = (n & 0x000000ff) << 24u;
	b1 = (n & 0x0000ff00) << 8u;
	b2 = (n & 0x00ff0000) >> 8u;
	b3 = (n & 0xff000000) >> 24u;

	return (b0 | b1 | b2 | b3);

}

void rotate_right(int width, int height, unsigned int *input, unsigned int *output)
{
	for (int r = 0; r < width; r++)
	{
		for (int c = 0; c < height; c++)
		{
			output[c * width + (width - r - 1)] = input[r + width * c];
		}
	}
}

void transpose(int width, int height, unsigned int *input, unsigned int *output)
{
	for (int r = 0; r < width; r++)
	{
		for (int c = 0; c < height; c++)
		{
			output[c + (height - 1 - r) * width] = input[r + c * width];
		}
	}
}

void rotate_right(int width, int height, unsigned char *input, unsigned char *output)
{
	for (int r = 0; r < width; r++)
	{
		for (int c = 0; c < height; c++)
		{
			output[c * width + (width - r - 1)] = input[r + width * c];
		}
	}
}

void transpose(int width, int height, unsigned char *input, unsigned char *output)
{
	for (int r = 0; r < width; r++)
	{
		for (int c = 0; c < height; c++)
		{
			output[c + (height - 1 - r) * width] = input[r + c * width];
		}
	}
}


int mapping[] = {
	48,
	49,
	50,
	51,
	52,
	53,
	54,
	55,
	56,
	57,
	65,
	66,
	67,
	68,
	69,
	70,
	71,
	72,
	73,
	74,
	75,
	76,
	77,
	78,
	79,
	80,
	81,
	82,
	83,
	84,
	85,
	86,
	87,
	88,
	89,
	90,
	97,
	98,
	100,
	101,
	102,
	103,
	104,
	110,
	113,
	114,
	116
};

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
	PSTR szCmdLine, int iCmdShow)
{
	static TCHAR szAppName[] = TEXT("NeuralNetwork");
	HWND         hwnd;
	MSG          msg;
	WNDCLASS     wndclass;

	wndclass.style = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc = WndProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = hInstance;
	wndclass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wndclass.lpszMenuName = NULL;
	wndclass.lpszClassName = szAppName;

	if (!RegisterClass(&wndclass))
	{
		MessageBox(NULL, TEXT("Program requires Windows NT!"),
			szAppName, MB_ICONERROR);
		return 0;
	}

	hwnd = CreateWindow(szAppName, TEXT("Neural Network Number Recognition"),
		WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, 200,
		CW_USEDEFAULT, 250,
		NULL, NULL, hInstance, NULL);

	ShowWindow(hwnd, iCmdShow);
	UpdateWindow(hwnd);

	while (GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return msg.wParam;
}

void draw_pixels(HDC hdc, HDC hdcMem, int width, int height)
{
	HBITMAP hBitmap, hOldBitmap;

	hBitmap = CreateCompatibleBitmap(hdc, width, height);
	SetBitmapBits(hBitmap, sizeof(int) * width * height, vram);
	hdcMem = CreateCompatibleDC(hdc);
	hOldBitmap = (HBITMAP)SelectObject(hdcMem, hBitmap);

	// This scaling is a little strange because Stretch maintains aspect ratios
	StretchBlt(hdc, 0, 0, cxClient, cyClient, hdcMem, 0, 0, width, height, SRCCOPY);
	SelectObject(hdcMem, hOldBitmap);
	DeleteDC(hdcMem);
	DeleteObject(hBitmap);
}



void load_random()
{
	int sample = rand() % MAX_TRAIN_IMAGE;

	unsigned int size;

	uint8_t* train_image =  (uint8_t*)get_file("data/emnist-balanced-train-images-idx3-ubyte", &size);
	uint8_t* train_labels = (uint8_t*)get_file("data/emnist-balanced-train-labels-idx1-ubyte", &size);

	uint32_t case_count = byteswap_uint32(*(uint32_t*)(train_image + 4));

	for (int i = sample - 1; i < sample; i++)
	{
		uint8_t* img = train_image + 16 + i * (MNIST_IMG_WIDTH * MNIST_IMG_HEIGHT);
		uint8_t* label = train_labels + 8 + i;
		uint8_t temp[MNIST_IMG_WIDTH * MNIST_IMG_HEIGHT];

		rotate_right(MNIST_IMG_WIDTH, MNIST_IMG_HEIGHT, img, temp);
		transpose(MNIST_IMG_WIDTH, MNIST_IMG_HEIGHT, temp, img);

		for (int y = 0; y < MNIST_IMG_HEIGHT; y++)
		{
			for (int x = 0; x < MNIST_IMG_WIDTH; x++)
			{
				vram[x + y * MNIST_IMG_WIDTH].r = ~img[x + y * MNIST_IMG_WIDTH];
				vram[x + y * MNIST_IMG_WIDTH].g = ~img[x + y * MNIST_IMG_WIDTH];
				vram[x + y * MNIST_IMG_WIDTH].b = ~img[x + y * MNIST_IMG_WIDTH];
			}
		}
	}

	delete[] train_image;
	delete[] train_labels;
}


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static RECT	rect = { 0, 0, 1, 1 };
	static HDC	hdcMem;
	PAINTSTRUCT ps;
	HDC hdc;
	static int	old_style = 0;
	static int	new_style = WS_CHILD | WS_VISIBLE;
	static int	xres, yres;
	static int guess = -1;
	static vector<layer_t*> layers;
	static float results[NUMBER_OF_OUTPUT_CELLS] = { 0 };
	static bool loaded = false;



	switch (message)
	{
	case WM_TIMER:
	{
		int width = cxClient;
		int height = cyClient;

		if (!loaded)
		{
			//		printf("Training Neural Network\r\n");
			train(layers);
			//		printf("Done\r\n");
			loaded = true;
		}

		InvalidateRect(hwnd, &rect, FALSE);
		return 0;
	}
	case WM_CREATE:
	{
		memset(vram, ~0, MNIST_IMG_WIDTH * MNIST_IMG_HEIGHT * sizeof(int));
		xres = MNIST_IMG_WIDTH;
		yres = MNIST_IMG_HEIGHT;

		SetWindowPos(hwnd, 0, 200, 200, 200, 250, 0);
		SetTimer(hwnd, WM_TICK, 16, NULL);
		return 0;
	}
	case WM_SIZE:
		cxClient = LOWORD(lParam);
		cyClient = HIWORD(lParam);
		rect.right = cxClient;
		rect.bottom = cyClient;
		return 0;

	case WM_KEYDOWN:
		switch (wParam)
		{
		case VK_SPACE:
			clear_pixels();
			break;
		case VK_RETURN:
		{
			uint8_t vbyte[MNIST_IMG_WIDTH * MNIST_IMG_HEIGHT];

			for (int i = 0; i < MNIST_IMG_WIDTH * MNIST_IMG_HEIGHT; i++)
			{
				rgb_t *rgb = (rgb_t *)&vram[i];

				vbyte[i] = (unsigned char)(rgb->r);
				vbyte[i] = ~vbyte[i];
			}

			guess = -1;
			guess = test(layers, vbyte, results);
			break;
		}
		case VK_TAB:
			load_random();
			break;

		}
		return 0;

	case WM_MOUSEMOVE:

		if (wParam & MK_LBUTTON)
		{
			float x = (float)LOWORD(lParam) / cxClient;
			float y = (float)HIWORD(lParam) / cyClient;

			rgb_t rgb;
			rgb_t black = { 0 };

			rgb.r = 64;
			rgb.g = 64;
			rgb.b = 64;

			set_pixel((int)(x * xres), (int)(y * yres), black); // center
			add_pixel((int)(x * xres + 1), (int)(y * yres), rgb); // right
			add_pixel((int)(x * xres - 1), (int)(y * yres), rgb);   // left
			add_pixel((int)(x * xres), (int)(y * yres + 1), rgb); // up
			add_pixel((int)(x * xres), (int)(y * yres - 1), rgb); // down
			add_pixel((int)(x * xres + 1), (int)(y * yres + 1), rgb); // up right
			add_pixel((int)(x * xres - 1), (int)(y * yres - 1), rgb); // down left
			add_pixel((int)(x * xres - 1), (int)(y * yres + 1), rgb); // up left
			add_pixel((int)(x * xres + 1), (int)(y * yres - 1), rgb); // down right
		}

		return 0;

	case WM_PAINT:
	{
		char guess_str[512];

		if (loaded)
		{
			if (guess != -1)
				sprintf(guess_str, "results: %c %3.1f confidence", mapping[guess], results[guess]);
			else
				sprintf(guess_str, "results: ? 0.0 confidence");
		}
		else
		{
			sprintf(guess_str, "Loading Training data");
		}

		hdc = BeginPaint(hwnd, &ps);
		draw_pixels(hdc, hdcMem, xres, yres);
		DrawText(hdc, guess_str, -1, &rect, 0);
		EndPaint(hwnd, &ps);
		return 0;
	}
	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	}
	return DefWindowProc(hwnd, message, wParam, lParam);
}


char *get_file(char *filename, unsigned int *size)
{
	FILE	*file;
	char	*buffer;
	int	file_size, bytes_read;

	file = fopen(filename, "rb");
	if (file == NULL)
		return 0;
	fseek(file, 0, SEEK_END);
	file_size = ftell(file);
	fseek(file, 0, SEEK_SET);
	buffer = new char[file_size + 1];
	bytes_read = (int)fread(buffer, sizeof(char), file_size, file);
	if (bytes_read != file_size)
	{
		delete[] buffer;
		fclose(file);
		return 0;
	}
	fclose(file);
	buffer[file_size] = '\0';

	if (size != NULL)
	{
		*size = file_size;
	}

	return buffer;
}

uint32_t byteswap_uint32(uint32_t a)
{
	return ((((a >> 24) & 0xff) << 0) |
		(((a >> 16) & 0xff) << 8) |
		(((a >> 8) & 0xff) << 16) |
		(((a >> 0) & 0xff) << 24));
}