#include "include.h"

using namespace std;

float train(vector<layer_t*>& layers, tensor_t<float>& data, tensor_t<float>& expected)
{
	for (unsigned int i = 0; i < layers.size(); i++)
	{
		if (i == 0)
			activate(layers[i], data);
		else
			activate(layers[i], layers[i - 1]->out);
	}

	tensor_t<float> grads = layers.back()->out - expected;

	for (int i = layers.size() - 1; i >= 0; i--)
	{
		if (i == layers.size() - 1)
			calc_grads(layers[i], grads);
		else
			calc_grads(layers[i], layers[i + 1]->grads_in);
	}

	for (unsigned int i = 0; i < layers.size(); i++)
	{
		fix_weights(layers[i]);
	}

	float err = 0;
	for (int i = 0; i < grads.size.x * grads.size.y * grads.size.z; i++)
	{
		float f = expected.data[i];
		if (f > 0.5)
			err += abs(grads.data[i]);
	}
	return err * 100;
}


void forward(vector<layer_t*>& layers, tensor_t<float>& data)
{
	for (unsigned int i = 0; i < layers.size(); i++)
	{
		if (i == 0)
			activate(layers[i], data);
		else
			activate(layers[i], layers[i - 1]->out);
	}
}

struct case_t
{
	tensor_t<float> data;
	tensor_t<float> out;
};




vector<case_t> read_test_cases()
{
	vector<case_t> cases;

	unsigned int size = 0;
	uint8_t* train_image = (uint8_t*)get_file("data/emnist-balanced-train-images-idx3-ubyte", &size);
	uint8_t* train_labels = (uint8_t*)get_file("data/emnist-balanced-train-labels-idx1-ubyte", &size);

	uint32_t case_count = byteswap_uint32(*(uint32_t*)(train_image + 4));

	for (unsigned int i = 0; i < case_count; i++)
	{
		case_t c{ tensor_t<float>(MNIST_IMG_WIDTH, MNIST_IMG_HEIGHT, 1), tensor_t<float>(NUMBER_OF_OUTPUT_CELLS, 1, 1) };

		uint8_t* img = train_image + 16 + i * (MNIST_IMG_WIDTH * MNIST_IMG_HEIGHT);
		uint8_t* label = train_labels + 8 + i;

		uint8_t temp[MNIST_IMG_WIDTH * MNIST_IMG_HEIGHT];


		rotate_right(MNIST_IMG_WIDTH, MNIST_IMG_HEIGHT, img, temp);
		transpose(MNIST_IMG_WIDTH, MNIST_IMG_HEIGHT, temp, img);


		for (int x = 0; x < MNIST_IMG_WIDTH; x++)
			for (int y = 0; y < MNIST_IMG_HEIGHT; y++)
				c.data(x, y, 0) = img[x + y * MNIST_IMG_WIDTH] / 255.f;

		for (int b = 0; b < NUMBER_OF_OUTPUT_CELLS; b++)
			c.out(b, 0, 0) = *label == b ? 1.0f : 0.0f;

		cases.push_back(c);
	}
	delete[] train_image;
	delete[] train_labels;

	return cases;
}

int train(vector<layer_t*> &layers)
{
	vector<case_t> cases = read_test_cases();

	
	// layout one
	conv_layer_t * layer1 = new conv_layer_t(1, 5, 8, cases[0].data.size);		// 28 * 28 * 1 -> 24 * 24 * 8 -- input pixels -> convolution layer
	relu_layer_t * layer2 = new relu_layer_t(layer1->out.size);					// convolution layer -> Rectified linear units (positive numbers only, throw away negatives)
	pool_layer_t * layer3 = new pool_layer_t(2, 2, layer2->out.size);			// 24 * 24 * 8 -> 12 * 12 * 8 -- pool layer, shrinks image down, max pooling (max of kernel box)
	fc_layer_t * layer4 = new fc_layer_t(layer3->out.size, NUMBER_OF_OUTPUT_CELLS);					// 4 * 4 * 16 -> 10 -- final classification maps pools to final 10 classes

	layers.push_back((layer_t*)layer1);
	layers.push_back((layer_t*)layer2);
	layers.push_back((layer_t*)layer3);
	layers.push_back((layer_t*)layer4);
	

	// layout two
	/*
	conv_layer_t * layer1 = new conv_layer_t(1, 5, 8, cases[0].data.size);		// 28 * 28 * 1 -> 24 * 24 * 8
	relu_layer_t * layer2 = new relu_layer_t(layer1->out.size);
	pool_layer_t * layer3 = new pool_layer_t(2, 2, layer2->out.size);				// 24 * 24 * 8 -> 12 * 12 * 8

	conv_layer_t * layer4 = new conv_layer_t(1, 3, 10, layer3->out.size);			// 12 * 12 * 6 -> 10 * 10 * 10
	relu_layer_t * layer5 = new relu_layer_t(layer4->out.size);
	pool_layer_t * layer6 = new pool_layer_t(2, 2, layer5->out.size);				// 10 * 10 * 10 -> 5 * 5 * 10

	fc_layer_t * layer7 = new fc_layer_t(layer6->out.size, 10);					// 4 * 4 * 16 -> 10

	layers.push_back((layer_t*)layer1);
	layers.push_back((layer_t*)layer2);
	layers.push_back((layer_t*)layer3);
	layers.push_back((layer_t*)layer4);
	layers.push_back((layer_t*)layer5);
	layers.push_back((layer_t*)layer6);
	layers.push_back((layer_t*)layer7);
	*/

	float amse = 0;
	int ic = 0;

	for (long ep = 0; ep < 100000; )
	{

		for (case_t& t : cases)
		{
			float xerr = train(layers, t.data, t.out);
			amse += xerr;

			ep++;
			ic++;

			if (ep % 1000 == 0)
			{
				printf("case %d err=%f\n", ep, amse / ic);
			}
		}
	}
	return 0;
}


int test(vector<layer_t*> &layers, uint8_t * data, float *results)
{
	uint8_t * usable = data;

	tensor_t<float> image(MNIST_IMG_WIDTH, MNIST_IMG_HEIGHT, 1);
	for (int i = 0; i < MNIST_IMG_HEIGHT; i++)
	{
		for (int j = 0; j < MNIST_IMG_WIDTH; j++)
		{
			image(j, i, 0) = (float)data[j + i * MNIST_IMG_WIDTH];
		}
	}

	forward(layers, image);
	tensor_t<float>& out = layers.back()->out;
	int max_index = -1;
	float max = FLT_MIN;
	for (int i = 0; i < NUMBER_OF_OUTPUT_CELLS; i++)
	{
		results[i] = out(i, 0, 0) * 100.0f;

		if (max < results[i])
		{
			max_index = i;
			max = results[i];
		}
	}
	return max_index;
}